# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('fill_like', 'getfields', 'setfields', 'any_isinstance',
           'copy_maybe', 'get_dct_values', 'get_obj_values',
           'get_instance', 'define_from', 'Chaperone')

class Chaperone:
    __slots__ = ('item',)
    def __init__(self, item):
        self.item = item
    def get_item(self, item):
        return self.item
    def set_item(self, item):
        self.item = item
    def __rerp__(self):
        return f"{type(self).__name__}(item={self.item})"

def fill_like(A, B, fields):
    setfield = setattr
    getfield = getattr
    for i in fields: setfield(A, i, getfield(B, i))
    
def getfields(obj, fields, getfield=getattr):
    return [getfield(obj, i) for i in fields]

def setfields(obj, names, fields, setfield=setattr):
    for i,j in zip(names, fields): setfield(obj, i, j)

def any_isinstance(objs, cls):
    isa = isinstance
    for i in objs:
        if isa(i, cls): return True
    return False

def copy_maybe(obj):
    return obj.copy() if hasattr(obj, 'copy') else obj

def get_dct_values(dct, params):
    return [dct[key] for key in params]

def get_obj_values(obj, params):
    attr = getattr
    return [attr(obj, key) for key in params]

def _define_from(cls, other, names):
    getfield = getattr
    setfield = setattr
    for name in names: setfield(cls, name, getfield(other, name))
    return cls

def define_from(other, names):
    return lambda cls: _define_from(cls, other, names)

def get_instance(iterable, cls):
    """Return object that is an instance of given class."""
    isa = isinstance
    objs = [i for i in iterable if isa(i, cls)]
    N = len(objs)
    if N == 1:
        return objs[0]
    elif N == 0:
        raise ValueError('instance not found')
    else:
        raise ValueError('multiple instances')


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import flexsolve as flx
from ..exceptions import InfeasibleRegion
from ..utils.decorators import thermo_user
from . import binary_phase_fraction as binary
from .dew_point import DewPointCache
from .bubble_point import BubblePointCache
from .fugacity_coefficients import IdealFugacityCoefficients
from .._thermal_condition import ThermalCondition
from ..indexer import MolarFlowIndexer
from .. import functional as fn
from ..utils import Cache
import numpy as np

__all__ = ('SLE',)

@thermo_user
class SLE:
    __slots__ = ('_x', # [float] Fraction of solute as a solid.
                 '_gamma', # [ActivityCoefficients] Estimates activity coefficients of a liquid.
                 '_liquid_mol', # [1d array] Liquid molar data.
                 '_solid_mol', # [1d array] Solid molar data.
                 '_phase_data', # tuple[str, 1d array] Phase-data pairs.
                 '_index', # [1d array] Index of chemicals in equilibrium.
                 '_chemical', # [Chemical] Single chemical in equilibrium.
                 '_nonzero', # [1d array(bool)] Chemicals present in the mixture
                 '_mol_solute', # [float] Solute molar data.
    )
    
    def __init__(self, imol=None, thermal_condition=None, thermo=None):
        super().__init__(imol, thermal_condition, thermo)
        imol = self._imol
        self._phase_data = tuple(imol)
        self._liquid_mol = liquid_mol = imol['l']
        self._solid_mol = imol['s']
        self._nonzero = np.zeros(liquid_mol.shape, dtype=bool)
        self._index = ()
        self._chemical = None
        self._x = None
    
    def _setup(self, solute):
        # Get flow rates
        liquid_mol = self._liquid_mol
        vapor_mol = self._vapor_mol
        mol = liquid_mol + vapor_mol
        nonzero = mol > 0
        chemicals = self.chemicals
        solute_index = chemicals.get_index(solute)
        self._mol_solute = mol_solute = nonzero[solute_index]
        if not mol_solute:
            raise RuntimeError('no solute available')
        if (self._nonzero == nonzero).all():
            index = self._index
        else:
            # Set up indices for both equilibrium and non-equilibrium species
            index = chemicals.get_lle_indices(nonzero)   
            N = len(index)
            if N < 2:
                raise RuntimeError('at least 2 chemicals are required for SLE')
            else:
                # Set equilibrium objects
                eq_chems = chemicals.tuple
                eq_chems = [eq_chems[i] for i in index]
                self._nonzero = nonzero
                self._index = index
                thermo = self._thermo
                self._gamma = thermo.Gamma(eq_chems)
        if mol.sum() != 0: 
            raise RuntimeError('no chemicals to perform equilibrium')
        self._mol = mol[index]
        
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

__all__ = ('all_same_chemicals',
           'all_same_phases',
           'same_chemicals',
           'same_phases',
)

def chemicals_match(chemical_users, chemicals):
    return all([chemicals is i.chemicals for i in chemical_users if i])

def phases_match(multiphase_users, phases):
    return all([phases == i.phases for i in multiphase_users if i])

def all_same_chemicals(obj, others):
    chemicals = obj.chemicals
    assert chemicals_match(others, chemicals), "chemicals do not match"
    
def all_same_phases(obj, others):
    phases = obj.phases
    assert phases_match(others, phases), "phases do not match"
    
def same_chemicals(obj, other):
    assert obj.chemicals is other.chemicals, "chemicals do not match"
    
def same_phases(obj, other):
    assert obj.phases == other.phases, "phases do not match"


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('repr_IDs_data', 'repr_kwargs', 'repr_kwarg', 'repr_couples',
           'repr_listed_values')
    
def repr_IDs_data(IDs, data, dlim=", ", start=None):
    return (start or dlim) + dlim.join([f"{ID}={i:.4g}" for ID, i in zip(IDs, data) if i])

def repr_kwargs(kwargs, dlim=", ", start=None):
    if kwargs:
        start = dlim if start is None else ""
        return start + dlim.join([f"{key}={value}" for key, value in kwargs.items()])
    else:
        return ""

def repr_args(args, dlim=", ", start=None):
    if args:
        return (start or dlim) + dlim.join([repr(value) for value in args])
    else:
        return ""

def repr_kwarg(name, value, spec=None, dlim=", "):
    if value:
        if spec: value = format(value, spec)
        else: value = repr(value)
        return dlim + f"{name}={value}"
    else:
        return ""

def repr_couples(dlim, IDs, data):
    return dlim.join([f"('{ID}', {i:.4g})" for ID, i in zip(IDs, data) if i])

def repr_listed_values(values):
    *values, last = values 
    if values:
        return ", ".join(values) + ' and ' + last
    else:
        return last

def repr_obj(obj):
    return f"{type(obj).__name__}({repr_kwargs(obj.__dict__, start='')})"







# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import equilibrium as eq
from ._chemical import Chemical
from ._chemicals import Chemicals
from .mixture import Mixture, ideal_mixture
from .utils import read_only, cucumber
from ._settings import settings

__all__ = ('Thermo',)


@cucumber # Just means you can pickle it
@read_only
class Thermo:
    """
    Create a Thermo object that defines a thermodynamic property package
    
    Parameters
    ----------
    chemicals : Chemicals or Iterable[str]
        Pure component chemical data.
    mixture : Mixture, optional
        Calculates mixture properties.
    Gamma : ActivityCoefficients subclass, optional
        Class for computing activity coefficients.
    Phi : FugacityCoefficients subclass, optional
        Class for computing fugacity coefficients.
    PCF : PoyntingCorrectionFactor subclass, optional
        Class for computing poynting correction factors.
    
    Examples
    --------
    >>> from thermosteam import Thermo
    >>> Thermo(['Ethanol', 'Water'])
    Thermo(chemicals=CompiledChemicals([Ethanol, Water]), mixture=Mixture(rule='ideal mixing', ..., include_excess_energies=False), Gamma=DortmundActivityCoefficients, Phi=IdealFugacityCoefficients, PCF=IdealPoyintingCorrectionFactors)
    
    Attributes
    ----------
    chemicals : Chemicals or Iterable[str]
        Pure component chemical data.
    mixture : Mixture, optional
        Calculates mixture properties.
    Gamma : ActivityCoefficients subclass, optional
        Class for computing activity coefficients.
    Phi : FugacityCoefficients subclass, optional
        Class for computing fugacity coefficients.
    PCF : PoyntingCorrectionFactor subclass, optional
        Class for computing poynting correction factors.
    
    """
    __slots__ = ('chemicals', 'mixture', 'Gamma', 'Phi', 'PCF',
                 'ideal_equilibrium_thermo') 
    
    def __init__(self, chemicals, mixture=None,
                 Gamma=eq.DortmundActivityCoefficients,
                 Phi=eq.IdealFugacityCoefficients,
                 PCF=eq.IdealPoyintingCorrectionFactors):
        if not isinstance(chemicals, Chemicals): chemicals = Chemicals(chemicals)
        if mixture:
            assert isinstance(mixture, Mixture), (
                f"mixture must be a '{Mixture.__name__}' object")
        else:
            mixture = ideal_mixture(chemicals)
        chemicals.compile()
        if settings._debug:
            issubtype = issubclass
            assert issubtype(Gamma, eq.ActivityCoefficients), (
                f"Gamma must be a '{eq.ActivityCoefficients.__name__}' subclass")
            assert issubtype(Phi, eq.FugacityCoefficients), (
                f"Phi must be a '{eq.FugacityCoefficients.__name__}' subclass")
            assert issubtype(PCF, eq.PoyintingCorrectionFactors), (
                f"PCF must be a '{eq.PoyintingCorrectionFactors.__name__}' subclass")
        
        setattr = object.__setattr__
        if (Gamma is eq.IdealActivityCoefficients
            and Phi is eq.IdealFugacityCoefficients
            and PCF is eq.IdealPoyintingCorrectionFactors):
            ideal = self
        else:
            cls = self.__class__
            ideal = cls.__new__(cls)
            setattr(ideal, 'chemicals', chemicals)
            setattr(ideal, 'mixture', mixture)
            setattr(ideal, 'Gamma', eq.IdealActivityCoefficients)
            setattr(ideal, 'Phi', eq.IdealFugacityCoefficients)
            setattr(ideal, 'PCF', eq.IdealPoyintingCorrectionFactors)
            setattr(ideal, 'ideal_equilibrium_thermo', ideal)
        setattr(self, 'chemicals', chemicals)
        setattr(self, 'mixture', mixture)
        setattr(self, 'Gamma', Gamma)
        setattr(self, 'Phi', Phi)
        setattr(self, 'PCF', PCF)
        setattr(self, 'ideal_equilibrium_thermo', ideal)
    
    @property
    def equilibrium_model(self):
        if self.ideal_equilibrium_thermo is self:
            return "Raoult's law"
        elif self.Gamma in (eq.UNIFACActivityCoefficients, eq.DortmundActivityCoefficients):
            return "modified Raoult's law"
        else:
            return "unknown"
    
    def as_chemical(self, chemical):
        isa = isinstance
        if isa(chemical, str):
            try: 
                chemical = self.chemicals.retrieve(chemical)
            except:
                chemical = Chemical(chemical)
        elif not isa(chemical, Chemical):
            raise ValueError('can only convert string to chemical')
        return chemical
    
    def subgroup(self, IDs):
        chemicals = self.chemicals.subgroup(IDs)
        return type(self)(chemicals, None, self.Gamma, self.Phi, self.PCF)
    
    def __repr__(self):
        return f"{type(self).__name__}(chemicals={self.chemicals}, mixture={self.mixture}, Gamma={self.Gamma.__name__}, Phi={self.Phi.__name__}, PCF={self.PCF.__name__})"
    
    def show(self):
        try:
            mixture_info = self.mixture._info().replace('\n', '\n    ')
        except:
            mixture_info = str(self.mixture)
        print(f"{type(self).__name__}(\n"
              f"    chemicals={self.chemicals},\n"
              f"    mixture={mixture_info},\n"
              f"    Gamma={self.Gamma.__name__},\n"
              f"    Phi={self.Phi.__name__},\n"
              f"    PCF={self.PCF.__name__}\n"
               ")")
    _ipython_display_ = show